#include <stdlib.h>
#include <string.h>
#include "tree.h"

/* how much to make text grow by */
#define TEXT_GROW 32

node *root;

/* recursive function to load a branch of the tree */
void load_nest(FILE *fp, node *top);
/* load text from document until a tag or EOF */
char *load_text(FILE *fp);
/* read text from the stream.  stop at tag or on EOF */
char *read_text(FILE *in);
/* read a tag from the document */
enum tag_type read_tag(FILE *fp, attribs *attr);

void load(FILE *fp)
{
	char *text;	/* text for a new text node */
	node *tmp;	/* tmp node space */

	root = (node *) malloc(sizeof(node));

	root->is_tag = 1;
	root->tag    = TAG_DUMMY;
	root->text   = NULL;

	root->attr.bold       = 0;
	root->attr.emphasis   = 0;
	root->attr.italic     = 0;
	root->attr.strong     = 0;
	root->attr.typewriter = 0;
	root->attr.underline  = 0;
	root->attr.size       = 10;
	root->attr.color      = COLOR_DEFAULT;
	
	root->child_count = 0;
	root->parent = NULL;

	/* breaks on EOF in middle of loop */
	while(1)
	{
		/* try to get text */
		text = read_text(fp);
		
		/* if we got text, make a node */
		if(text)
		{
			tmp = (node *) malloc(sizeof(node));
			tmp->is_tag      = 0;
			tmp->tag         = TAG_DUMMY; /******************/
			tmp->text        = text;
			memcpy(&tmp->attr, &root->attr, sizeof(attribs));
			tmp->child_count = 0;         /******************/
			tmp->parent      = root;
			
			/* add to children */
			root->children[root->child_count] = tmp;
			root->child_count += 1;
		}
		
		/* if we got no text, it's either a tag, or EOF */
		if(feof(fp)) break;
		
		/* create a tag node, and recursively fill it in */
		tmp = (node *) malloc(sizeof(node));
		tmp->is_tag = 1;
		tmp->text   = NULL;	/********************/
		tmp->parent      = root;

		/* modify attributes and tag id for specific tag */		
		memcpy(&tmp->attr, &root->attr, sizeof(attribs));
		tmp->tag = read_tag(fp, &tmp->attr);
		
		/********quick check - should never happen on valid input***********/
		if(tmp->tag == TAG_END)
			fprintf(stderr, "invalid input: tag closing where no open tag\n");
		
		/* load nested tags and text */
		load_nest(fp, tmp);
	
		/* add to children */
		root->children[root->child_count] = tmp;
		root->child_count += 1;
		
	}
}

void load_nest(FILE *fp, node *top)
{
	char   *text;	/* text for a new text node */
	node   *tmp;	/* tmp node space */
	int     tag;	/* tmp storage space for tag id */
	attribs attr;	/* tmp storage space for some attributes */

	top->child_count = 0;

	/* breaks on EOF in middle of loop */
	while(1)
	{
		/* try to get text */
		text = read_text(fp);
		
		/* if we got text, make a node */
		if(text)
		{
			tmp = (node *) malloc(sizeof(node));
			tmp->is_tag      = 0;
			tmp->tag         = TAG_DUMMY; /******************/
			tmp->text        = text;
			memcpy(&tmp->attr, &top->attr, sizeof(attribs));
			tmp->child_count = 0;         /******************/
			tmp->parent      = top;
			
			/* add to children */
			top->children[top->child_count] = tmp;
			top->child_count += 1;
		}
		
		/* if we got no text, it's either a tag, or EOF */
		if(feof(fp)) break;
		
		/* get some info about the tag */
		memcpy(&attr, &top->attr, sizeof(attr));
		tag = read_tag(fp, &attr);
		
		/* if it's an end tag, this branch is done */
		if(tag == TAG_END) return;
		
		/* create tag node */
		tmp = (node *) malloc(sizeof(node));
		tmp->is_tag = 1;
		tmp->tag    = tag;
		tmp->text   = NULL;	/********************/
		memcpy(&tmp->attr, &attr, sizeof(attribs));
		tmp->parent      = top;

		/* load nested tags and text */
		load_nest(fp, tmp);
	
		/* add to children */
		top->children[top->child_count] = tmp;
		top->child_count += 1;
		
	}
}

char *read_text(FILE *in)
{
	char *text;	/* the text that's been read          */
	char *tmp;	/* temp space use to extend _text_    */
	int   input;	/* used to read file 1 char at a time */

	text = NULL;	/* defensive programming */

	/* read a character */
	input = getc(in);

	/* keep reading until EOF of the '<' from a tag */
	while(input != EOF && input != '<')
	{
		/* add character to _text_ */
		if(text != NULL)	/* append to current string */
		{
			/* don't calculate strlen() over and over */
			int len = strlen(text);

			/* check: does it need to grow?                  */
			/* the two adds the new char and NULL terminator */
			if((len + 2) % TEXT_GROW == 1)
			{
				/* might not be pretty, but ought to work */
				tmp = (char *) malloc(sizeof(char) * TEXT_GROW * (((len + 1) / TEXT_GROW) + 1));

				/* copy what we already have */
				strcpy(tmp, text);

				/* free previous _text_ and set it to _tmp_ */
				free(text);
				text = tmp;
			}

			/* whether it grew or not, append _input_ */
			text[len]     = (char) input;
			text[len + 1] = '\0';
		}
		else			/* start string from scratch */
		{
			text = (char *) malloc(sizeof(char) *  TEXT_GROW);
			text[0] = (char) input;
			text[1] = '\0';
		}

		/* read a new character */
		input = (char) getc(in);
	}

	/* return the fruit of our efforts */
	return text;
}

enum tag_type read_tag(FILE *fp, attribs *attr)
{
	int tag;
	
	tag = getc(fp);
	getc(fp);	/* read off ">" for most tags */
		
	switch(tag)
	{
		case 'B':	/* bold */
			if(attr->bold) return TAG_DUMMY;
			attr->bold = 1;
			return TAG_BOLD;
		case 'E':	/* emphasis */
			getc(fp);				/* read off ">" */
			
			/* if strong is set, tell it its's a dummy tag */
			if(!attr->strong)
			{
				attr->emphasis ^= 1;	/* invert, don't set */
				return TAG_EMPHASIS;
			}
			else return TAG_DUMMY;
		case 'I':  	/* italic */
			if(attr->italic) return TAG_DUMMY;
			attr->italic = 1;
			return TAG_ITALIC;
		case 'P':	/* plain */
			getc(fp);				/* read off ">" */
			/* if everything is already as plain would make it, ignore */
			if(!attr->underline && !attr->bold && !attr->emphasis
				&& !attr->italic && !attr->strong && !attr->typewriter)
					return TAG_DUMMY;
			attr->underline  = 0;	/* set a bunch of attributes to off */
			attr->bold       = 0;
			attr->emphasis   = 0;
			attr->italic     = 0;
			attr->strong     = 0;
			attr->typewriter = 0;
			return TAG_PLAIN;
		case 'S':	/* strong */
			if(attr->strong) return TAG_DUMMY;
			attr->strong = 1;
			return TAG_STRONG;
		case 'T':	/* typewriter */
			getc(fp);				/* read off ">" */
			if(attr->typewriter) return TAG_DUMMY;
			attr->typewriter = 1;
			return TAG_TYPEWRITER;
		case 'U':	/* underline */
			if(attr->underline < 3)		/* only has effect if underline < 3 */
			{
				attr->underline += 1;
				return TAG_UNDERLINE;
			}
			else return TAG_DUMMY;
		case '0':	/* size */
			if(attr->size == 0) return TAG_DUMMY;
			attr->size = 0;
			return TAG_SIZE_0;
		case '1':
			if(attr->size == 1) return TAG_DUMMY;
			attr->size = 1;
			return TAG_SIZE_1;
		case '2':
			if(attr->size == 2) return TAG_DUMMY;
			attr->size = 2;
			return TAG_SIZE_2;
		case '3':
			if(attr->size == 3) return TAG_DUMMY;
			attr->size = 3;
			return TAG_SIZE_3;
		case '4':
			if(attr->size == 4) return TAG_DUMMY;
			attr->size = 4;
			return TAG_SIZE_4;
		case '5':
			if(attr->size == 5) return TAG_DUMMY;
			attr->size = 5;
			return TAG_SIZE_5;
		case '6':
			if(attr->size == 6) return TAG_DUMMY;
			attr->size = 6;
			return TAG_SIZE_6;
		case '7':
			if(attr->size == 7) return TAG_DUMMY;
			attr->size = 7;
			return TAG_SIZE_7;
		case '8':
			if(attr->size == 8) return TAG_DUMMY;
			attr->size = 8;
			return TAG_SIZE_8;
		case '9':
			if(attr->size == 9) return TAG_DUMMY;
			attr->size = 9;
			return TAG_SIZE_9;
		case 'r':	/* color */
			if(attr->color == COLOR_RED) return TAG_DUMMY;
			attr->color = COLOR_RED;
			return TAG_COLOR_RED;
		case 'g':
			if(attr->color == COLOR_GREEN) return TAG_DUMMY;
			attr->color = COLOR_GREEN;
			return TAG_COLOR_GREEN;
		case 'b':
			if(attr->color == COLOR_BLUE) return TAG_DUMMY;
			attr->color = COLOR_BLUE;
			return TAG_COLOR_BLUE;
		case 'c':
			if(attr->color == COLOR_CYAN) return TAG_DUMMY;
			attr->color = COLOR_CYAN;
			return TAG_COLOR_CYAN;
		case 'm':
			if(attr->color == COLOR_MAGENTA) return TAG_DUMMY;
			attr->color = COLOR_MAGENTA;
			return TAG_COLOR_MAGENTA;
		case 'y':
			if(attr->color == COLOR_YELLOW) return TAG_DUMMY;
			attr->color = COLOR_YELLOW;
			return TAG_COLOR_YELLOW;
		case 'k':
			if(attr->color == COLOR_BLACK) return TAG_DUMMY;
			attr->color = COLOR_BLACK;
			return TAG_COLOR_BLACK;
		case 'w':
			if(attr->color == COLOR_WHITE) return TAG_DUMMY;
			attr->color = COLOR_WHITE;
			return TAG_COLOR_WHITE;
		case '/':
			if(getc(fp) != '>') getc(fp);	/* read off remaining tag chars */
			return TAG_END;
		default:
			fprintf(stderr, "unknown first tag character: \'%c\'\n", tag);
			return TAG_DUMMY;
	}
}

